#include <fcntl.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>

#include "config.h"
#include "libhttpd.h"


httpd_server* httpd_initialize( httpd_sockaddr* sa, int port )
{
	httpd_server* hs;

	hs = NEW( httpd_server, 1 );
	if ( hs == (httpd_server*) 0 )
		return (httpd_server*) 0;

	hs->port = port;

	if ( sa == (httpd_sockaddr*) 0 )
		hs->listen_fd = -1;
	else
		hs->listen_fd = initialize_listen_socket( sa );

	if ( hs->listen_fd == -1)
	{
		free_httpd_server( hs );
		return (httpd_server*) 0;
	}

	return hs;
}


int initialize_listen_socket( httpd_sockaddr* sa )
{
	int listen_fd;
	int on, flags;

	listen_fd = socket( sa->sa.sa_family, SOCK_STREAM, 0 );
	if ( listen_fd < 0 )
		return -1;
	(void) fcntl( listen_fd, F_SETFD, 1 );
	(void) fcntl( listen_fd, F_SETFL, O_NONBLOCK);
	on = 1;
	(void) setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (char*) &on,sizeof(on) );
	if ( bind( listen_fd, &sa->sa, sizeof(struct sockaddr_in) ) < 0 )
	{
		(void) close( listen_fd );
		return -1;
	}
	flags = fcntl( listen_fd, F_GETFL, 0 );
	if ( flags == -1 )
	{
		(void) close( listen_fd );
		return -1;
	}
	if ( fcntl( listen_fd, F_SETFL, flags | O_NDELAY ) < 0 )
	{
		(void) close( listen_fd );
		return -1;
	}
	if ( listen( listen_fd, LISTEN_BACKLOG ) < 0 )
	{
		(void) close( listen_fd );
		return -1;
	}
	return listen_fd;
}


void free_httpd_server( httpd_server* hs )
{
	hs = (httpd_server*) 0;
	free( (void*) hs );
}


void clear_client( connecttab* connects, int cnum, int* numconnects )
{
	connects[cnum].conn_state = CNST_FREE;
	if ( connects[cnum].hc->conn_fd > 0 )
		close( connects[cnum].hc->conn_fd );
	connects[cnum].hc->conn_fd = 0;
	(*numconnects)--;
	conn_init( connects[cnum].hc );
}


int httpd_get_conn( int listen_fd, connecttab* connects, int cnum )
{
	httpd_sockaddr sa;
	int sz, sockfd;

	sz = sizeof(sa);
	sockfd = accept( listen_fd, &sa.sa, &sz );
	if ( sockfd < 0 )
		return -1;

	conn_init( connects[cnum].hc );
	connects[cnum].hc->conn_fd = sockfd;
	time( &( connects[cnum].hc->now ) );
	(void) fcntl( connects[cnum].hc->conn_fd, F_SETFD, 1 );
	memset( &(connects[cnum].hc->client_addr), 0, sizeof(connects[cnum].hc->client_addr) );
	memcpy( &(connects[cnum].hc->client_addr), &sa, sizeof(struct sockaddr_in) );

	return sockfd;
}


void conn_first_init( httpd_conn* hc )
{
	hc->hostname = NEW( char, HC_DEF_BUF );
	hc->method = NEW( char, HC_METHOD_BUF );
	hc->encodedurl = NEW( char, HC_URL_BUF );
	hc->decodedurl = NEW( char, ( HC_URL_BUF * 2 ) );
	hc->origfilename = NEW( char, HC_FILENAME_BUF );
	hc->referer = NEW( char, HC_URL_BUF );
	hc->read_buf = NEW( char, RECV_BUF_SIZE );
	hc->accept = NEW( char, HC_DEF_BUF );
	hc->accepte = NEW( char, HC_DEF_BUF );
	hc->encodings = NEW( char, HC_DEF_BUF );
	hc->protocol = NEW( char, HC_PROTOCOL_BUF );
	hc->query = NEW( char, HC_URL_BUF );
	hc->useragent = NEW( char, HC_DEF_BUF );
	hc->cookie = NEW( char, HC_URL_BUF );
	hc->contenttype = NEW( char, HC_DEF_BUF );
	hc->keep_alive = hc->one_one = hc->sendstart = hc->contentlength = 0;
	hc->status = STAT_NOTFOUND;
	hc->send_idx = hc->read_idx = hc->read_size = hc->file = hc->conn_fd = 0;
}


void conn_init( httpd_conn* hc )
{
	if ( hc->file > 0 )
		close( hc->file );
	hc->file = 0;

	if ( hc->read_size > RECV_BUF_SIZE )
	{
		free( hc->read_buf );
		hc->read_buf = NEW( char, RECV_BUF_SIZE );
	}
	memset( hc->read_buf, 0, RECV_BUF_SIZE );
	memset( hc->hostname, 0, HC_DEF_BUF );
	memset( hc->method, 0, HC_METHOD_BUF );
	memset( hc->encodedurl, 0, HC_URL_BUF );
	memset( hc->decodedurl, 0, ( HC_URL_BUF * 2 ) );
	memset( hc->origfilename, 0, HC_FILENAME_BUF );
	memset( hc->referer, 0, HC_URL_BUF );
	memset( hc->accept, 0, HC_DEF_BUF );
	memset( hc->accepte, 0, HC_DEF_BUF );
	memset( hc->encodings, 0, HC_DEF_BUF );
	memset( hc->protocol, 0, HC_PROTOCOL_BUF );
	memset( hc->query, 0, HC_URL_BUF );
	memset( hc->useragent, 0, HC_DEF_BUF );
	memset( hc->cookie, 0, HC_URL_BUF );
	memset( hc->contenttype, 0, HC_DEF_BUF );
	hc->keep_alive = hc->one_one = hc->sendstart = hc->contentlength = 0;
	hc->status = STAT_NOTFOUND;
	hc->send_idx = hc->read_idx = hc->read_size = 0;
}

int httpd_got_request( httpd_conn* hc )
{
	int request_recv;
	if ( hc->read_size >= RECV_BUF_SIZE )
		hc->read_buf = RENEW( hc->read_buf, char, hc->read_size + RECV_BUF_SIZE );
	request_recv = read( hc->conn_fd, &(hc->read_buf[hc->read_size]), RECV_BUF_SIZE );
	if ( request_recv < 0 )
	{
		free( hc->read_buf );
		hc->read_buf = NEW( char, RECV_BUF_SIZE );
		return -1;
	}
	hc->read_size += request_recv;
	if ( hc->read_size > RECV_MAX_BUF_SIZE )
		return -1;
	if ( request_recv > 0 )
		time( &( hc->now ) );
	if ( strcmp( &hc->read_buf[( hc->read_size - 4 )], "\r\n\r\n" ) == 0 || strcmp( &hc->read_buf[( hc->read_size - 2 )], "\n\n" ) == 0 )
	{
		if ( httpd_parse_request( hc ) < 0 )
			return -1;
		return 1;
	}else
		return 0;
}


char* getbuf( httpd_conn* hc )
{
	int i;
	char c;

	for ( i = hc->read_idx; hc->read_idx < hc->read_size; ++hc->read_idx )
	{
		c = hc->read_buf[hc->read_idx];
		if ( c == '\n' || c == '\r' )
		{
			hc->read_buf[hc->read_idx] = '\0';
			++hc->read_idx;
			if ( c == '\r' && hc->read_idx < hc->read_size && hc->read_buf[hc->read_idx] == '\n' )
			{
				hc->read_buf[hc->read_idx] = '\0';
				++hc->read_idx;
			}
			return &(hc->read_buf[i]);
		}
	}
	return (char*) 0;
}


int httpd_parse_request( httpd_conn* hc )
{
	char* buf;
	char* method_str;
	char* url;
	char* decodedurl;
	char* protocol;
	char* hostname;
	char* dummy;
	char* filename;

	method_str = getbuf( hc );
	if ( method_str == (char*) 0 )
	{
		//httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
		return -1;
	}

	if ( strncasecmp( method_str, "GET", 3 ) == 0 )
		strncpy( hc->method, "GET", 3 );
	else if ( strncasecmp( method_str, "POST", 4 ) == 0 )
		strncpy( hc->method, "POST", 4 );
	else if ( strncasecmp( method_str, "HEAD", 4 ) == 0 )
		strncpy( hc->method, "HEAD", 4 );
	else{
		//httpd_send_err( hc, 501, err501title, "", err501form, method_str );
		return -1;
	}
	url = (char*) strtok( &method_str[strlen( hc->method )], " " );
	protocol = (char*) strtok( NULL, " " );

	if ( protocol == (char*) 0 )
	{
		strncpy( hc->protocol, "HTTP/0.9", 8 );
	}else{
		if ( strcasecmp( protocol, "HTTP/1.0" ) != 0 )
			hc->one_one = 1;
		strncpy( hc->protocol, protocol, 8 );
	}

	if  ( strncasecmp( url, "http://", 7 ) == 0 )
	{
		if ( ! hc->one_one )
		{
			//httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
			return -1;
		}
		hostname = url + 7;
		url = (char*) strchr( hostname, '/' );
		if ( url == (char*) 0 )
		{
			//httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
			return -1;
		}
		*url = '\0';
		strncpy( hc->hostname, hostname, HC_DEF_BUF );
	}

	strncpy( hc->encodedurl, url, HC_URL_BUF );
	decodedurl = NEW( char, ( HC_URL_BUF * 2 ) );

	dummy = (char*) strchr( hc->encodedurl, '?' );
	if ( dummy != (char*) 0 )
	{
		dummy++;
		strncpy( hc->query, dummy, HC_URL_BUF );
		memset( &hc->encodedurl[( strlen(hc->encodedurl) - strlen(dummy) - 1 )], 0, ( HC_URL_BUF - strlen( hc->encodedurl ) - strlen( dummy ) ) );
	}

	strdecode( decodedurl, hc->encodedurl );
	de_dotdot( hc->decodedurl );

	if ( decodedurl[0] != '/' || decodedurl[1] == '/' || ( decodedurl[1] == '.' && decodedurl[2] == '.' && ( decodedurl[3] == '\0' || decodedurl[3] == '/' ) ) )
	{
		free( decodedurl );
		//httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
		return -1;
	}

	strncpy( hc->decodedurl, decodedurl, ( HC_URL_BUF * 2 ) );
	free( decodedurl );
	strncpy( hc->origfilename, hc->decodedurl, HC_FILENAME_BUF );

	if ( hc->origfilename[( strlen( hc->origfilename ) - 1 )] == '/' )
		hc->origfilename = (char*) strcat( hc->origfilename, INDEXFILE );

	while ( ( buf = getbuf( hc ) ) != (char*) 0 )
	{
		if ( buf[0] == '\0' )
			break;
		if ( strncasecmp( buf, "Referer:", 8 ) == 0 )
		{
			dummy = &buf[8];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->referer, dummy, HC_URL_BUF );
		}else if ( strncasecmp( buf, "User-Agent:", 11 ) == 0 ) {
			dummy = &buf[11];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->useragent, dummy, HC_DEF_BUF );
		}else if ( strncasecmp( buf, "Host:", 5 ) == 0 ) {
			dummy = &buf[5];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->hostname, dummy, HC_DEF_BUF );
			dummy = (char*) strchr( hc->hostname, ':' );
			if ( dummy != (char*) 0 )
				*dummy = '\0';
		}else if ( strncasecmp( buf, "Accept:", 7 ) == 0 ) {
			dummy = &buf[7];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->accept, dummy, HC_DEF_BUF );
		}else if ( strncasecmp( buf, "Accept-Encoding:", 16 ) == 0 ) {
			dummy = &buf[16];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->accepte, dummy, HC_DEF_BUF );
	    }else if ( strncasecmp( buf, "If-Modified-Since:", 18 ) == 0 ) {
			//dummy = &buf[18];
			//hc->if_modified_since = tdate_parse( cp );
			//if ( hc->if_modified_since == (time_t) -1 )
				//syslog( LOG_DEBUG, "unparsable time: %.80s", cp );
		}else if ( strncasecmp( buf, "Cookie:", 7 ) == 0 ) {
			dummy = &buf[7];
			dummy += strspn( dummy, " \t" );
			strncpy( hc->cookie, dummy, HC_URL_BUF );
		}else if ( strncasecmp( buf, "Authorization:", 14 ) == 0 ) {
			//dummy = &buf[14];
			//dummy += strspn( dummy, " \t" );
			//hc->authorization = dummy;
		}else if ( strncasecmp( buf, "Content-Type:", 13 ) == 0 ) {
			dummy = &buf[13];
			dummy += strspn( dummy, " \t" );
			hc->contenttype = (char*) atoi( dummy );
		}else if ( strncasecmp( buf, "Content-Length:", 15 ) == 0 ) {
			dummy = &buf[15];
			hc->contentlength = atol( dummy );
		}else if ( strncasecmp( buf, "Connection:", 11 ) == 0 && hc->one_one == 1 ) {
			dummy = &buf[11];
			dummy += strspn( dummy, " \t" );
			if ( strcasecmp( dummy, "keep-alive" ) == 0 )
				hc->keep_alive = 1;
		}
	}
	return 0;
}


void strdecode( char* to, char* from )
{
	for ( ; *from != '\0'; ++to, ++from )
	{
		if ( from[0] == '%' && isxdigit( from[1] ) && isxdigit( from[2] ) )
		{
			*to = hexit( from[1] ) * 16 + hexit( from[2] );
			from += 2;
		}else{
			*to = *from;
		}
	}
	*to = '\0';
}


int hexit( char c )
{
	if ( c >= '0' && c <= '9' )
		return c - '0';
	if ( c >= 'a' && c <= 'f' )
		return c - 'a' + 10;
	if ( c >= 'A' && c <= 'F' )
		return c - 'A' + 10;
	return 0;
}


void de_dotdot( char* file )
{
	char* cp;
	char* cp2;
	int l;

	while ( ( cp = (char*) strstr( file, "//") ) != (char*) 0 )
	{
		for ( cp2 = cp + 2; *cp2 == '/'; ++cp2 )
			continue;
		(void) strcpy( cp + 1, cp2 );
	}

	while ( ( cp = (char*) strstr( file, "/../" ) ) != (char*) 0 )
	{
		for ( cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2 )
			continue;
		if ( cp2 < file )
			break;
		(void) strcpy( cp2, cp + 3 );
	}

	while ( ( l = strlen( file ) ) > 3 && strcmp( ( cp = file + l - 3 ), "/.." ) == 0 )
	{
		for ( cp2 = cp - 1; cp2 >= file && *cp2 != '/'; --cp2 )
			continue;
		if ( cp2 < file )
			break;
		*cp2 = '\0';
	}
}


int httpd_start_request( httpd_conn* hc )
{
	int sendlen, result;
	char* sendbuf;
	char* filename;
	int len;

	if ( hc == (httpd_conn*) 0 ) return -1;

	if ( hc->sendstart == 0 )
	{
		filename = NEW( char, ( HC_FILENAME_BUF + strlen( DOCUMENT_ROOT ) ) );
		(void) sprintf( filename, "%s%s", DOCUMENT_ROOT, hc->origfilename );
		if ( lstat( filename, &hc->filestat ) != 0 )
			return -1;
		hc->file = open( filename, O_RDONLY );
		if ( hc->file < 0 )
			return -1;
		hc->status = STAT_OK;
		hc->sendstart = 1;
		sendbuf = make_header( hc );
		result = send( hc->conn_fd, sendbuf, strlen( sendbuf ), 0);
		free( filename );
		free( sendbuf );
		if ( result < 0 )
			return -1;
	}
	sendbuf = NEW( char, SEND_BUF_SIZE );
	if ( hc->send_idx != 0 )
	{
		if ( lseek( hc->file, hc->send_idx, SEEK_SET ) < 0 )
			return -1;
	}
	result = read( hc->file, sendbuf, SEND_BUF_SIZE );
	if ( result < 0 )
	{
		free( sendbuf );
		return -1;
	}
	sendlen = write( hc->conn_fd, sendbuf, result );
	if ( sendlen <= 0 )
		return -1;
	hc->send_idx += sendlen;
	free( sendbuf );
	if ( hc->send_idx == (long int) hc->filestat.st_size )
	{
		return 1;
	}
	return 0;
}


char* make_header( httpd_conn* hc )
{
	char* header;
	char* mdate;
	char* statr;

	header = NEW( char, (int) ( strlen( hc->protocol ) + strlen( SERVER_SOFTWARE ) + 128 ) );
	mdate = (char*) ctime( &(hc->filestat.st_mtime) );
	mdate[( strlen( mdate )-1 )] = '\0';
	statr = status_code( hc->status );
	sprintf( header, "%s %d %s\r\nServer: %s\r\nContent-type: text/html\r\nLast-Modified: %s\r\nContent-Length: %d\r\n\r\n", hc->protocol, hc->status, statr, SERVER_SOFTWARE, mdate, hc->filestat.st_size );

	return header;
}


char *status_code( int status )
{
	switch ( status )
	{
		case STAT_OK : return STATR_200; break;
		case STAT_BAD : return STATR_400; break;
		case STAT_NOAUTH : return STATR_401; break;
		case STAT_FORBIDDEN : return STATR_403; break;
		case STAT_NOTFOUND : return STATR_404; break;
	}
}

